#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "common.h"
#include "oppd_conf.h"

LOCAL_FUNC void on_call_return_new( int select );
LOCAL_FUNC void on_call_return_overwrite_confirm( int select );
LOCAL_FUNC void plist_to_clist(conf_t *plist, const int num);
LOCAL_FUNC void on_call_return_main( int select );
LOCAL_FUNC void edit_start( edit_mode mode );
LOCAL_FUNC void on_call_return_edit( int select );
LOCAL_FUNC void on_call_return ( int select );
LOCAL_FUNC void info_off( void );

int     Number_of_printers = 0;

/*
 * MAIN WINDOW
 */

void
on_button_exit_released                (GtkButton       *button,
                                        gpointer         user_data)
{
        if( is_changed_plist( Plist, Number_of_printers ) )
        {
                info_on( FROM_MAIN, NULL );
        } else {
                on_call_return_main( OK );
        }
}

LOCAL_FUNC void on_call_return_main( int select )
{
        if ( OK == select )
        {
                gtk_main_quit();
        } else {
                ;          /* Main window close cancel, do nothing. */
        }
}

FUNCTION void review_main_button( void )
{
        GtkWidget *widget;

        ASSERT( NULL != gp->window_main );

        widget = lookup_widget(gp->window_main, "button_list_edit");
        if( !Number_of_printers )
                gtk_widget_set_sensitive(widget, FALSE);
        else
                gtk_widget_set_sensitive(widget, TRUE);

        widget = lookup_widget(gp->window_main, "button_list_del");
        if( !Number_of_printers )
                gtk_widget_set_sensitive(widget, FALSE);
        else
                gtk_widget_set_sensitive(widget, TRUE);

        widget = lookup_widget(gp->window_main, "button_save");
        if( (!Number_of_printers) || is_changed_plist ( Plist, Number_of_printers ) )
                gtk_widget_set_sensitive(widget, TRUE);
        else
                gtk_widget_set_sensitive(widget, FALSE);

        widget = lookup_widget(gp->window_main, "button_list_edit");
        if( !Number_of_printers)
                gtk_widget_set_sensitive(widget, FALSE);
        else
                gtk_widget_set_sensitive(widget, TRUE);
}

/*
 * Printer List control.
 */
void
on_button_list_add_released            (GtkButton       *button,
                                        gpointer         user_data)
{
        edit_start( MODE_ADD );
}

void
on_button_list_edit_released           (GtkButton       *button,
                                        gpointer         user_data)
{
        edit_start( MODE_EDIT );
}


void
on_button_list_del_released            (GtkButton       *button,
                                        gpointer         user_data)
{
        GtkWidget *widget;

        if(!Number_of_printers)
                return;

        gp->mode = MODE_DEL;

        widget = lookup_widget(GTK_WIDGET(button), "clist_printer");
        gp->focus_printer = GTK_CLIST(widget)->focus_row;

        Number_of_printers = del_plist( Plist, Number_of_printers, gp->focus_printer );

        plist_to_clist(Plist, Number_of_printers);
}

/*
 * File control.
 */
void
on_button_file_new_released            (GtkButton       *button,
                                        gpointer         user_data)
{
        gp->mode = MODE_NEW;

        if( is_changed_plist ( Plist, Number_of_printers ) )
                info_on( FROM_NEW, NULL );
        else
                on_call_return_new( OK );
}

LOCAL_FUNC void on_call_return_new( int select )
{
        if ( OK == select )
        {
                GtkWidget *widget;
                widget = lookup_widget(GTK_WIDGET(gp->window_main), "clist_printer");
                gtk_clist_clear (GTK_CLIST(widget));
                Number_of_printers = 0;
                review_main_button();
        } else {
                ;          /* all clear cancel, do nothing. */
        }
}

void
on_button_load_released                (GtkButton       *button,
                                        gpointer         user_data)
{
        gp->mode = MODE_LOAD;

        if( is_changed_plist ( Plist, Number_of_printers ) )
                info_on( FROM_LOAD, NULL );
        else
                on_call_return_load( OK );
}

FUNCTION void on_call_return_load( int select )
{
        if ( OK == select )
        {
                Number_of_printers = read_conf( Plist, OPPD_CONF );
                if( Number_of_printers > 0 )
                {
                        plist_to_clist(Plist, Number_of_printers);
                        clear_all_changed_flag(Plist, Number_of_printers);
                }
        } else {
                Number_of_printers = 0; /* load oppd.conf cancel, do nothing. */
        }
}

void
on_button_save_released                (GtkButton       *button,
                                        gpointer         user_data)
{
        gp->mode = MODE_SAVE;

        if( !Number_of_printers )
        {
                info_on( FROM_OVERWRITE, _("No printer entry.\nIf press OK, force overwrite empty oppd.conf.\nOverwrite?") );
        } else {
                on_call_return_overwrite_confirm( OK );
        }
}

LOCAL_FUNC void on_call_return_overwrite_confirm( int select )
{
        if ( OK == select )
        {
                if( 0 == save_conf(Plist, OPPD_CONF, Number_of_printers) )
                        clear_all_changed_flag(Plist, Number_of_printers);
                else
                        info_on( FROM_SAVE_ERR, _("!error: oppd.conf could not saved.\nPlease restart me, on root user.") );
        } else {
                ;          /* save oppd.conf cancel, do nothing. */
        }
	review_main_button();
}


LOCAL_FUNC void plist_to_clist(conf_t *plist, const int num)
{
        GtkWidget *widget;
        GtkCList *clist;
        const char *clist_null_data[] = { "", "", "", "" };

        dbgp(dbg,"* start\n");

        widget = lookup_widget(GTK_WIDGET(gp->window_main), "clist_printer");
        clist = GTK_CLIST(widget);

        /* start freeze list view display. */
        gtk_clist_freeze (clist);

        /* clear all clist data. */
        gtk_clist_clear (clist);

        {
                int i;

                /* cleate clist */
                for (i = 0; i < num; i++ ) {
                        gtk_clist_append (clist, (char **)clist_null_data );
                        gtk_clist_set_text (clist, i, 0, plist[i].name);
                        gtk_clist_set_text (clist, i, 1, plist[i].device);
                        gtk_clist_set_text (clist, i, 2, plist[i].module);
                        gtk_clist_set_text (clist, i, 3, plist[i].description);
                }
        }

        dbgp(dbg,"mode=%d, focus_printer=%d\n", gp->mode, gp->focus_printer);
        if( MODE_DEL == gp->mode )
        {
                ASSERT( 0 <= Number_of_printers );

                if( 0 > gp->focus_printer) {
                        gp->focus_printer = 0;
                }
                else if ( gp->focus_printer >= Number_of_printers ) {
                        gp->focus_printer = Number_of_printers -1;
                }

        }
        else if( MODE_ADD == gp->mode )
        {
                gp->focus_printer = Number_of_printers -1;
        }

        if( (MODE_NEW  != gp->mode) &&
            (MODE_LOAD != gp->mode) &&
            (0 != Number_of_printers) ) {
                gtk_clist_select_row (clist, gp->focus_printer, 0);
                clist->focus_row = gp->focus_printer;
        }

        /* unfreeze list view display. */
        gtk_clist_thaw (clist);
        review_main_button();
}


/*
 * PRINTER CONFIGULATION WINDOW
 */
LOCAL_FUNC void edit_start( edit_mode mode )
{
        GtkWidget *widget;

        gp->mode = mode;
        gp->edit_dialog = create_edit_dialog ();

        widget = GTK_WIDGET(gp->window_main);
        widget = lookup_widget(widget, "clist_printer");

        if ( MODE_ADD  == mode )
                gp->focus_printer = Number_of_printers;

        if ( MODE_EDIT == mode ) {
                GtkWidget *edit;

                gp->focus_printer = (GTK_CLIST(widget))->focus_row;

                edit = GTK_WIDGET(gp->edit_dialog);
                widget = lookup_widget(edit, "printer_name_value");
                gtk_entry_set_text((GtkEntry *)widget,
                                   Plist[gp->focus_printer].name);

                widget = lookup_widget(edit, "description_value");
                gtk_entry_set_text((GtkEntry *)widget,
                                   Plist[gp->focus_printer].description);

                if( !Plist[gp->focus_printer].location ){

                        /* Local printer */

                        widget = lookup_widget(edit, "radiobutton_local");
                        gtk_toggle_button_set_active((GtkToggleButton *)widget, TRUE);

/*
                        widget = lookup_widget(edit, "device_value");
                        gtk_entry_set_text((GtkEntry *)widget,
                                           Plist[gp->focus_printer].device);

                        widget = lookup_widget(edit, "module_value");
                        gtk_entry_set_text((GtkEntry *)widget,
                                           Plist[gp->focus_printer].module);
*/

                } else {

                        /* Network printer */

                        widget = lookup_widget(edit, "radiobutton_network");
                        gtk_toggle_button_set_active((GtkToggleButton *)widget, TRUE);

                        widget = lookup_widget(edit, "server_value");
                        gtk_entry_set_text((GtkEntry *)widget,
                                           Plist[gp->focus_printer].device);

                        widget = lookup_widget(edit, "remote_printer_value");
                        gtk_entry_set_text((GtkEntry *)widget,
                                           Plist[gp->focus_printer].module);

                }
        }

        gtk_widget_show (gp->edit_dialog);
}

/*
 * Printer location
 */
void
on_radiobutton_local_toggled           (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
        GtkWidget *widget;
        widget = lookup_widget ( GTK_WIDGET ( togglebutton ), "frame_local" );

        if ( TRUE == gtk_toggle_button_get_active ( togglebutton ) )
        {
                gtk_widget_show ( widget );
        } else {
                gtk_widget_hide ( widget );
        }
}


void
on_radiobutton_network_toggled         (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
        GtkWidget *widget;
        widget = lookup_widget ( GTK_WIDGET ( togglebutton ), "frame_network" );

        if ( TRUE == gtk_toggle_button_get_active ( togglebutton ) )
        {
                gtk_widget_show ( widget );
        } else {
                gtk_widget_hide ( widget );
        }
}


void
on_edit_cancelbutton_released          (GtkButton       *button,
                                        gpointer         user_data)
{
        on_call_return_edit( OK );
}

void
on_edit_okbutton_released              (GtkButton       *button,
                                        gpointer         user_data)
{
        GtkEntry  *entry;
        GtkWidget *widget;
        GtkWidget *edit = GTK_WIDGET(gp->edit_dialog);
        const char *str;

        conf_t conf = {
                1,         /* int  is_set; */
                0,         /* int  location; */
                0,         /* int  changed; */
                { '\0', }, /* char name       [MAX_PRINTER_NAME +1]; */
                { '\0', }, /* char device     [MAX_DEVICE_NAME  +1]; */
                { '\0', }, /* char module     [MAX_MODULE_NAME  +1]; */
                { '\0', }  /* char description[MAX_DESCRIPTION  +1]; */
        };

        entry = (GtkEntry *)lookup_widget(edit, "printer_name_value");
        str = gtk_entry_get_text(entry);

        if( !is_str(str) )
        {
                info_on( FROM_EDIT_NONAME, _("Printer name is must recommending item.\nPrease specify printer name.") );
                return;
        }
        if( batting_name(Plist, Number_of_printers, str, gp->focus_printer) ){
                info_on( FROM_EDIT_BATTING, _("The printer name is already exist.\nPrease change the name, to another unique name.") );
                return;
        }

        strncpy(conf.name, str, MAX_PRINTER_NAME);

        entry = (GtkEntry *)lookup_widget(edit, "description_value");
        strncpy(conf.description, gtk_entry_get_text(entry), MAX_DESCRIPTION);

        widget = lookup_widget(edit, "radiobutton_local");
        if ( TRUE == gtk_toggle_button_get_active((GtkToggleButton *)widget) )
        {
                entry = (GtkEntry *)lookup_widget(edit, "device_value");
                strncpy(conf.device, gtk_entry_get_text(entry), MAX_DEVICE_NAME);

                entry = (GtkEntry *)lookup_widget(edit, "module_value");
                strncpy(conf.module, gtk_entry_get_text(entry), MAX_MODULE_NAME);
        }
        else                    /* Network */
        {
                conf.location = 1;

                entry = (GtkEntry *)lookup_widget(edit, "server_value");
                str = gtk_entry_get_text(entry);
                if( !is_str(str) )
                {
                        info_on( FROM_EDIT_NONAME, _("Print server name is must recommending item.\nPrease specify print server name.") );
                        return;
                }
                strcat(conf.device, str);

                if( NULL == strchr(conf.device, ':') )
                        strcat(conf.device, ":22222");

                entry = (GtkEntry *)lookup_widget(edit, "remote_printer_value");
                str = gtk_entry_get_text(entry);
                if( !is_str(str) )
                {
                        info_on( FROM_EDIT_NONAME, _("Remote printer name is must recommending item.\nPrease specify remote printer name.") );
                        return;
                }
                strncpy(conf.module, str, MAX_MODULE_NAME);
        }

	if( 0 > Number_of_printers )
	  Number_of_printers = 0;

	if( 0 > gp->focus_printer )
	  gp->focus_printer = 0;

        Number_of_printers = put_plist( Plist, Number_of_printers, &conf, gp->focus_printer );

        plist_to_clist(Plist, Number_of_printers);

        on_call_return_edit ( OK );
}

LOCAL_FUNC void on_call_return_edit( int select )
{
        if ( OK == select )
        {
                gtk_widget_destroy (gp->edit_dialog);
                gp->edit_dialog = 0;
        } else {
                ;          /* Edit window close cancel, do nothing. */
        }
        gp->mode = MODE_NONE;
}

/*
 * User select OK or Cancel.
 */
void
on_judge_cancelbutton_released         (GtkButton       *button,
                                        gpointer         user_data)
{
        on_call_return ( CANCEL );
}

void
on_judge_okbutton_released             (GtkButton       *button,
                                        gpointer         user_data)
{
        on_call_return ( OK );
}

LOCAL_FUNC void on_call_return ( int select )
{
        if (      FROM_MAIN == gp->from )
        {
                on_call_return_main ( select );
        }
        else if ( FROM_LOAD == gp->from )
        {
                on_call_return_load ( select );
        }
        else if ( FROM_EDIT == gp->from )
        {
                on_call_return_edit ( select );
        }
        else if ( FROM_EDIT_NONAME == gp->from )
        {
                ;               /* do nothing */
        }
        else if ( FROM_EDIT_BATTING == gp->from )
        {
                ;               /* do nothing */
        }
        else if ( FROM_NOT_ROOT == gp->from )
        {
		if ( CANCEL == select )
			gtk_main_quit();
        }
        else if ( FROM_SAVE_ERR == gp->from )
        {
                ;               /* do nothing */
        }
        else if ( FROM_OVERWRITE == gp->from )
        {
                on_call_return_overwrite_confirm ( select );
        }
        else
        {
                dbgp(dbg,"!error: on_call == FROM_NONE\n");
        }
        info_off ();
}

FUNCTION void info_on( on_call from, const char *msg )
{
        GtkWidget *widget;

        gp->from = from;
        gp->info_dialog = create_info_dialog ();

        if( NULL != msg )
        {
                widget = lookup_widget(GTK_WIDGET(gp->info_dialog), "info_lavel");
                gtk_label_set_text((GtkLabel *)widget, msg);
        }
        gtk_widget_show (gp->info_dialog);
}

LOCAL_FUNC void info_off( void )
{
        gtk_widget_destroy(gp->info_dialog);
        gp->info_dialog = 0;
        gp->from = FROM_NONE;
}

/* end of file. */
