#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "oppd_conf.h"

conf_t _plist[MAX_PR_LIST] = {
        { 0, },
};

conf_t *Plist = _plist;

#define MAX_READ_BUFF 1023

LOCAL_FUNC int set_plist ( conf_t *plist, const char *buff, const int size );
LOCAL_FUNC void copy_conf( conf_t *plist, const conf_t *conf );
LOCAL_FUNC int  cmp_conf( const conf_t *conf1, const conf_t *conf2 );

/** read oppd.conf
 *
 * @retval  -1 file read error.
 * @retval int number of printers. If zore is no printer entry.
 */
FUNCTION int read_conf ( conf_t *plist, const char *filepath )
{
        FILE *fp;
        int n = 0;
        int i = 0;
        char buff[MAX_READ_BUFF +1];

        dbgp(dbg,"* start\n");

        fp = fopen(filepath, "r");

        if( !fp )
                return -1;
        {
                char c;

                for( c = fgetc(fp); (EOF != c) && ('\0' != c) && (i < MAX_READ_BUFF);
                     c = fgetc(fp) )
                {
                        buff[i] = c;
                        i++;
                }
        }
        fclose(fp);

        if( i == MAX_READ_BUFF )
        {
                fprintf(stderr, "%s:%d:%s: buffer over flow\n", dbg);
                exit(12);       /* No buffer space available */
        }
        buff[i] = '\0';

        n = set_plist( plist, buff, MAX_READ_BUFF );

        return n;
}

/** set printer list.
 *
 * @retval -1 file read error.
 * @retval int number of printers. If zore is no printer entry.
 */
LOCAL_FUNC int set_plist ( conf_t *plist, const char *buff, const int size )
{
        int n = 0;
        int i = 0;

        dbgp(dbg,"* start\n");

        while( '\0' != buff[i] )
        {
                int j;

                plist[n].is_set         =   0;
                plist[n].location       =   0;
                plist[n].changed        =   0;
                plist[n].name       [0] = '\0';
                plist[n].device     [0] = '\0';
                plist[n].module     [0] = '\0';
                plist[n].description[0] = '\0';

                while( ( ' ' == buff[i]) ||
                       ( '#' == buff[i]) ||
                       ('\r' == buff[i]) ||
                       ('\n' == buff[i]) )
                {
                        if( '#' == buff[i] )
                        {
                                while(('\r' != buff[i]) &&
                                      ('\n' != buff[i]) &&
                                      ('\0' != buff[i]) )
                                        i++;
                        } else {
                                i++;
                        }
                }

                for(j = 0;  ('\t' != buff[i]) &&
                            ('\r' != buff[i]) &&
                            ('\n' != buff[i]) &&
                            ('\0' != buff[i]) ; j++ )
                {
                        plist[n].name[j] = buff[i];
                        i++;
                }
                plist[n].name[j] = '\0';

                if( '\t' == buff[i])
                {
                        i++;
                        if( '/' == buff[i] )
                        {
                                plist[n].location = 0;

                                for(j = 0;  ('\t' != buff[i]) &&
                                            ('\r' != buff[i]) &&
                                            ('\n' != buff[i]) &&
                                            ('\0' != buff[i]) ; j++ )
                                {
                                        plist[n].device[j] = buff[i];
                                        i++;
                                }
                                plist[n].device[j] = '\0';

                                if('\t' == buff[i])
                                {
                                        i++;
                                        for(j = 0;  ('\t' != buff[i]) &&
                                                    ('\r' != buff[i]) &&
                                                    ('\n' != buff[i]) &&
                                                    ('\0' != buff[i]) ; j++ )
                                        {
                                                plist[n].module[j] = buff[i];
                                                i++;
                                        }
                                        plist[n].module[j] = '\0';

                                        if('\t' == buff[i])
                                        {
                                                i++;
                                                for(j = 0;  ('\r' != buff[i]) &&
                                                            ('\n' != buff[i]) &&
                                                            ('\0' != buff[i]) ; j++ )
                                                {
                                                        plist[n].description[j] = buff[i];
                                                        i++;
                                                }
                                                plist[n].description[j] = '\0';
                                        }
                                }
                        } else {
                                plist[n].location = 1;
                                i += strlen("oppd://");

                                for(j = 0;  ( '/' != buff[i]) &&
                                            ('\t' != buff[i]) &&
                                            ('\r' != buff[i]) &&
                                            ('\n' != buff[i]) &&
                                            ('\0' != buff[i]) ; j++ )
                                {
                                        plist[n].device[j] = buff[i];
                                        i++;
                                }
                                plist[n].device[j] = '\0';

                                if('/' == buff[i])
                                {
                                        i++;
                                        for(j = 0;  ('\t' != buff[i]) &&
                                                    ('\r' != buff[i]) &&
                                                    ('\n' != buff[i]) &&
                                                    ('\0' != buff[i]) ; j++ )
                                        {
                                                plist[n].module[j] = buff[i];
                                                i++;
                                        }
                                        plist[n].module[j] = '\0';
                                }

                                if('\t' == buff[i])
                                {
                                        i++;
                                        for(j = 0;  ('\r' != buff[i]) &&
                                                    ('\n' != buff[i]) &&
                                                    ('\0' != buff[i]) ; j++ )
                                        {
                                                plist[n].description[j] = buff[i];
                                                i++;
                                        }
                                        plist[n].description[j] = '\0';
                                }

                        }
                }

                if( '\0' != plist[n].name[0] )
                {
                        plist[n].is_set = 1;
                        n ++;
                }
        }

        return n;
}

FUNCTION int is_changed_plist ( const conf_t *plist, const int num_of_printers )
{
        int i;

        dbgp(dbg,"* start\n");

        if( !num_of_printers )
                return 0;

        ASSERT( NULL != plist );

        for( i = 0; i < num_of_printers; i++ )
        {
                if( plist[i].changed )
                        return 1;
        }
        return 0;
}

FUNCTION int del_plist ( conf_t *plist, const int num_of_printers, const int line )
{
        int i;

        dbgp(dbg,"* start\n");

        ASSERT( NULL != plist );
        ASSERT( 0 != num_of_printers );
        ASSERT( 0 <= line );
        ASSERT( num_of_printers > line );

        for( i = line; i < (num_of_printers -1); i++ )
        {
                copy_conf( &plist[i], &plist[i+1] );
                plist[i].changed  = 1;
        }
        memset(&plist[i], 0, sizeof(conf_t));

        return (num_of_printers -1);
}

FUNCTION int put_plist ( conf_t *plist, int num_of_printers, const conf_t *conf, const int line )
{
        int changed;

        dbgp(dbg,"* start\n");

        ASSERT( NULL != plist );
        ASSERT( NULL != conf );
        ASSERT( 0 <= num_of_printers  );
        ASSERT( 0 <= line  );
        ASSERT( line <= num_of_printers  );

        if( line == num_of_printers )
        {
                num_of_printers++;
                changed = 1;
        }
        else
        {
                changed = cmp_conf(&plist[line], conf);
        }

        copy_conf( &plist[line], conf );
        plist[line].changed = changed;

        return num_of_printers;
}

LOCAL_FUNC void copy_conf( conf_t *plist, const conf_t *conf )
{
        dbgp(dbg,"* start\n");

        ASSERT( NULL != plist );
        ASSERT( NULL != conf );

        memset(plist, 0, sizeof(conf_t));

        plist->is_set   =           conf->is_set;
        plist->location =           conf->location;
        plist->changed  =           conf->changed;
        memcpy( plist->name,        conf->name,       MAX_PRINTER_NAME);
        memcpy( plist->device,      conf->device,     MAX_DEVICE_NAME );
        memcpy( plist->module,      conf->module,     MAX_MODULE_NAME );
        memcpy( plist->description, conf->description,MAX_DESCRIPTION );

/*
        dbgp(dbg,"%d %d %d %s\t%s\t%s\t%s\n"
             , plist->is_set
             , plist->location
             , plist->changed
             , plist->name
             , plist->device
             , plist->module
             , plist->description
             );
*/
}

LOCAL_FUNC int  cmp_conf( const conf_t *conf1, const conf_t *conf2 )
{
        dbgp(dbg,"* start\n");

        ASSERT( NULL != conf1 );
        ASSERT( NULL != conf2 );

        if( (conf1->is_set   !=              conf2->is_set    ) ||
            (conf1->location !=              conf2->location  ) ||
            (conf1->changed  !=              conf2->changed   ) ||
            (0 != strcmp( conf1->name,       conf2->name )    ) ||
            (0 != strcmp( conf1->device,     conf2->device )  ) ||
            (0 != strcmp( conf1->module,     conf2->module )  ) ||
            (0 != strcmp( conf1->description,conf2->description)))
                return 1;
        return 0;
}

FUNCTION int save_conf ( const conf_t *plist, const char *filepath, const int num_of_printers )
{
        FILE *fp;
        int i;

        dbgp(dbg,"* start\n");

        ASSERT( NULL != plist );
        ASSERT( NULL != filepath );
        ASSERT( 0 <= num_of_printers );

        fp = fopen( filepath, "w" );
        if( !fp )
                return -1;

        for( i = 0; i < num_of_printers; i++ )
        {
                if( plist[i].is_set )
                {
                        if( !plist[i].location ) /* Local */
                                fprintf(fp, "%s\t%s\t%s", plist[i].name, plist[i].device, plist[i].module);
                        else                     /* Network */
                                fprintf(fp, "%s\toppd://%s/%s", plist[i].name, plist[i].device, plist[i].module);

                        if( *plist[i].description )
                                fprintf(fp, "\t%s", plist[i].description);

                        fprintf(fp, "\n");
                }
        }

        fclose(fp);
        return 0;
}

FUNCTION void clear_all_changed_flag ( conf_t *plist, const int num_of_printers )
{
        int i;

        dbgp(dbg,"* start\n");

        ASSERT( NULL != plist );
        ASSERT( 0 <= num_of_printers );

        for( i = 0; i < num_of_printers; i++ )
                plist[i].changed = 0;
}

FUNCTION int batting_name ( const conf_t *plist, const int num_of_printers, const char *name, const int line )
{
        int i;

        dbgp(dbg,"* start\n");

        ASSERT( NULL != plist );
        ASSERT( NULL != name );

        if( num_of_printers <= 0 )
                return 0;

        if( (line < 0) || (line >= num_of_printers) )
                return 0;

        for( i = 0; i < num_of_printers; i++ )
                if( ( i != line ) &&
                    ( 0 == strcmp(plist[i].name, name )) )
                        return 1;
        return 0;
}

/* end of file */
