/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
/*
 * oppd-listen.c
 * open-printing printer daemon listener
 */
#include <stdio.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <errno.h>

#include "oppd_common.h"
#include "oppd_conf.h"

typedef	struct {
	OPPD_ConfEnt	*conf;
	pid_t		pid;
	int		retry;
} dev_proc;

/*
 *
 */
oppd_signal	sig[] = {
	{SIGCHLD, NULL},
	{SIGTERM, NULL},
	{SIGPIPE, NULL},
	{SIGHUP , NULL},
	{-1     , NULL}
};
int		force_abort = 0;
int		force_restart = 0;
dev_proc	dev_proc_list[OPPD_CONF_MAX];

/*
 * signal handler
 */
static
void		sig_handler(
	int		signo
#ifdef USE_SIGINFO
	,
	siginfo_t	*siginfo,
	void		*ucntextptr
#endif
	)
{
	pid_t		pid;
	int		stat;
	int		i;

	switch (signo) {
		case	SIGCHLD :
			while((pid = waitpid(-1, &stat, WNOHANG)) > 0) {
				for (i = 0; i < OPPD_CONF_MAX; i++) {
					if (dev_proc_list[i].pid  == pid) {
						dev_proc_list[i].pid = 0;
						/*fprintf(stderr,
						  "listen:%s is dropped\n",
						  dev_proc_list[i].conf
						    ->printerName);*/
					}
				}
			}
			break;
		case	SIGALRM :
			/* ignore */
			break;
		case	SIGTERM :
			force_abort = 1;
			//fprintf(stderr,"listen:catch SIGTERM\n");
			break;
		case	SIGPIPE :
			/* ignore */
			break;
		case	SIGHUP :
			force_abort = 1;
			force_restart = 1;
			//fprintf(stderr,"listen:catch SIGHUP\n");
			break;
		default :
			/* ignore */
			break;
	}
	return;
}

/*
 * initialize signals
 */
static
int		sig_init(
	void
	)
{
	int		ecode = 0;
	int		i = 0;
struct	sigaction	act;
struct	sigaction	old;

	while (sig[i].signo != -1) {
		act.sa_flags = SA_RESTART;
#ifdef USE_SIGINFO
		act.sa_sigaction = sig_handler;
		act.sa_flags |= SA_SIGINFO;
#else
		act.sa_handler = sig_handler;
#endif
		sigemptyset(&act.sa_mask);
		if (sigaction(sig[i].signo, &act, &old) < 0) {
			ecode = -1;
			break;
		}
		sig[i].act = malloc(sizeof(struct sigaction));
		if (!sig[i].act) {
			ecode = -1;
			break;
		}
		memcpy(sig[i].act, &old, sizeof(struct sigaction));
		i++;
	}

	return ecode;
}

/*
 * finalize signals
 */
static
void		sig_fin(
	void
	)
{
	int		i = 0;

	while (sig[i].signo != -1) {
		if (!sig[i].act) {
			sigaction(sig[i].signo, sig[i].act, NULL);
			free(sig[i].act);
			sig[i].act = NULL;
		}
		i++;
	}

	return;
}

/*
 * open listener sockets
 */
static
int		open_sock(
	oppd_sock	*sock
	)
{
	int		ecode = 0;
	int		one = 1;
struct	sockaddr_in	addr;

	sock->print = socket(PF_INET, SOCK_STREAM, 0);
	if (setsockopt(sock->print,SOL_SOCKET,
	                SO_REUSEADDR,&one,sizeof(one)) < 0) {
		return -1;
	}
	fcntl(sock->print, F_SETFD, fcntl(sock->print, F_GETFD)|FD_CLOEXEC);
	if (sock->print != -1) {
		bzero(&addr, sizeof(struct sockaddr_in));
		addr.sin_family = PF_INET;
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
		addr.sin_port = htons(oppd_get_port(OPPD_SERV_NAME,
		                               "tcp", OPPD_PORT_DFLT));

		ecode = bind(sock->print,
		              (struct sockaddr *)&addr, sizeof(addr));
	} else {
		ecode = -1;
	}

	sock->status = socket(PF_INET, SOCK_STREAM, 0);
	if (setsockopt(sock->status,SOL_SOCKET,
	                SO_REUSEADDR,&one,sizeof(one)) < 0) {
		return -1;
	}
	fcntl(sock->status, F_SETFD, fcntl(sock->status, F_GETFD)|FD_CLOEXEC);
	if (sock->status != -1) {
		bzero(&addr, sizeof(struct sockaddr_in));
		addr.sin_family = PF_INET;
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
		addr.sin_port = htons(oppd_get_port(OPPS_SERV_NAME,
		                               "tcp", OPPS_PORT_DFLT));

		ecode = bind(sock->status,
		              (struct sockaddr *)&addr, sizeof(addr));
	} else {
		ecode = -1;
	}

	return ecode;
}

/*
 * close listener sockets
 */
static
void		close_sock(
	oppd_sock	*sock
	)
{
	if (sock->print != -1) {
		close(sock->print);
		sock->print = -1;
	}
	if (sock->status != -1) {
		close(sock->status);
		sock->status = -1;
	}

	return;
}

/*
 * initialize
 */
static
int		init(
	oppd_sock	*sock
	)
{
	int		ecode = 0;

	/* clear flags */
	force_abort = 0;
	force_restart = 0;

	/* initialize signal */
	if (!ecode) ecode = sig_init();

	/* set log-file */
	oppd_set_log();

 	/* open listener sockets */
	if (!ecode) ecode = open_sock(sock);

	return ecode;
}

/*
 * finalize
 */
static
void		fin(
	oppd_sock	*sock
	)
{
	int		i;

 	/* close listener sockets */
	close_sock(sock);

	/* finalize signals */
	sig_fin();

	/* terminate all process of this group */
	for (i = 0; i < OPPD_CONF_MAX; i++) {
		if (dev_proc_list[i].pid) {
			kill(dev_proc_list[i].pid, SIGHUP);
			dev_proc_list[i].conf = NULL;
			dev_proc_list[i].pid = 0;
		}
	}
	kill(0, SIGTERM);

	return;
}

/*
 * start process
 */
static
void		start_proc(
	char		*name,
	char		**argv
	)
{
	char		buf[1024];

	execvp(name, argv);
	strcpy(buf, "./");
	strcat(buf, name);
	execvp(buf, argv);
	exit(-1);
}

/*
 * start device process
 */
static
int		start_dev_proc(
	void
	)
{
	int		ecode = 0;
	OPPD_ConfEnt	*list;
	int		cnt;
	int		i;
	struct	stat	st;
	pid_t		pid;
	char		*arg[] = {OPPD_OUT_NAME,NULL,NULL};

	if (lstat(OPPD_SPL_DIR, &st) < 0) {
		if ((ecode = mkdir(OPPD_SPL_DIR, 0777))) return ecode;
	} else {
		if (!S_ISDIR(st.st_mode)) {
			ecode =- 1;
			return ecode;
		}
	}

	cnt = oppd_getConfList(OPPD_CONF_FILE,&list);

	for (i = 0; i < cnt; i++) {
		dev_proc_list[i].conf  = &(list[i]);
		dev_proc_list[i].pid   = 0;
		dev_proc_list[i].retry = 10;

		pid = fork();
		if (pid == 0) {
			arg[1] = list[i].printerName;
			start_proc(OPPD_OUT_NAME, arg);
			exit(-1);
		} else {
			dev_proc_list[i].pid = pid;
		}
	}
	while (i < OPPD_CONF_MAX) {
		dev_proc_list[i].conf  = NULL;
		dev_proc_list[i].pid   = 0;
		dev_proc_list[i].retry = 10;
		i++;
	}

	return ecode;
}

/*
 * re-start device process
 */
static
int		restart_dev_proc(
	void
	)
{
	int		ecode = 0;
	int		i;
	struct	stat	st;
	pid_t		pid;
	char		*arg[] = {OPPD_OUT_NAME,NULL,NULL};

	for (i = 0; i < OPPD_CONF_MAX; i++) {
		if ((dev_proc_list[i].conf) && (dev_proc_list[i].pid  == 0)) {
			dev_proc_list[i].retry--;
			if (dev_proc_list[i].retry > 0) {
				pid = fork();
				if (pid == 0) {
					arg[1]
					  = dev_proc_list[i].conf->printerName;
					start_proc(OPPD_OUT_NAME, arg);
					exit(-1);
				} else {
					dev_proc_list[i].pid = pid;
				}
			}
		}
	}

	return ecode;
}

/*
 * start spooler process
 */
static
int		start_spool_proc(
	int		sock_fd
	)
{
	int		ecode = 0;
	char		*arg[] = {OPPD_SPL_NAME,NULL};
	pid_t		pid;

	pid = fork();
	if (pid == 0) {
		dup2(sock_fd, 0);
		dup2(sock_fd, 1);
		fcntl(sock_fd, F_SETFD, fcntl(sock_fd, F_GETFD)|FD_CLOEXEC);

		start_proc(OPPD_SPL_NAME, arg);
	}

	return ecode;
}

/*
 * start status process
 */
static
int		start_stat_proc(
	int		sock_fd
	)
{
	int		ecode = 0;
	pid_t		pid;
	char		sfd[10];
	char		*arg[] = {OPPD_STA_NAME,"-s",sfd,NULL};

	sprintf(sfd,"%d",sock_fd);
	pid = fork();
	if (pid == 0) {
		start_proc(OPPD_STA_NAME, arg);
		exit(-1);
	}

	return ecode;
}

/*
 * listen sockets
 */
static
int		listen_sock(
	oppd_sock	*sock
	)
{
	int		ecode = 0;
	fd_set		fds;
	/*
struct	timeval		tv;
	*/
	int		sock_fd;
	int		max_fd;

	listen(sock->print, 5);
	listen(sock->status, 5);
	max_fd = (sock->print > sock->status ? sock->print : sock->status) + 1;

	while (!force_abort) {
		if (ecode) break;
		FD_ZERO(&fds);
		FD_SET(sock->print, &fds);
		FD_SET(sock->status, &fds);
		/* need not timeout
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		*/
		if (select(max_fd, &fds, NULL, NULL, NULL) < 0) {
			if (errno != EINTR) ecode = -1;
			continue;
		}

		/* re-start device process */
		restart_dev_proc();

		if (FD_ISSET(sock->print, &fds)) {
			if ((sock_fd = accept(sock->print,
						NULL,NULL)) < 0) {
				if (errno != EINTR) ecode = -1;
			}
			if (!ecode) ecode = start_spool_proc(sock_fd);
			close(sock_fd);
		}

		if (FD_ISSET(sock->status, &fds)) {
			if ((sock_fd = accept(sock->status,
			                      NULL,NULL)) < 0) {
				if (errno != EINTR) ecode = -1;
			}
			if (!ecode) ecode = start_stat_proc(sock_fd);
			close(sock_fd);
		}
	}

	return ecode;
}

/*
 * main
 */
int		main(
	int		argc,
	char		**argv
	)
{
	int		ecode = 0;
	oppd_sock	sock = {-1,-1};

	do {

		/* initialize */
		if (!ecode) ecode = init(&sock);

		/* start device process */
		if (!ecode) ecode = start_dev_proc();

		/* listen sockets */
		if (!ecode) ecode = listen_sock(&sock);

		/* finalize */
		fin(&sock);

	} while(force_restart);

	return ecode;
}

