/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
/*
 * oppd-spool.c
 * open-printing printer daemon spooler
 */
#include <stdio.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <time.h>
#include <cups.h>

#include "oppd_common.h"
#include "oppd_conf.h"

typedef	struct {
	char		tag;
	char		*name;
	int		len;
	char		*val;
} obj_attrib;
typedef	struct {
	char		tag;
	obj_attrib	attrib[10];
} ipp_attrib;

static	char		*buf = NULL;
static	int		buf_pos = 0;
static	char		printer[OPPD_BUF_LEN];

static	ipp_attrib	req_att[] = {
	{
		0x01,	/* operation-attributes-tag */
		{
			{
				0x47,	/* charset */
				"attributes-charset",
				-1,
				NULL
			},
			{
				0x48,	/* naturalLanguage */
				"attributes-natural-language",
				-1,
				NULL
			},
			{
				0x45,	/* uri */
				"printer-uri",
				-1,
				NULL
			},
			{
				0x42,	/* nameWithoutLanguage */
				"requesting-user-name",
				-1,
				NULL
			},
			{
				0x42,	/* nameWithoutLanguage */
				"job-name",
				-1,
				NULL
			},
			{
				0x49,	/* mimeMediaType */
				"document-format",
				-1,
				NULL
			},
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	},
	{
		0x02,	/* job-attributes-tag */
		{
			{
				0x21,	/* integer */
				"copies",
				-1,
				NULL,
			},
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	},
	{
		0x04,	/* printer-attributes-tag */
		{
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	},
	{
		0x03,	/* end-of-attributes-tag */
		{
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	}
};
static	ipp_attrib	rep_att[] = {
	{
		0x01,	/* operation-attributes-tag */
		{
			{
				0x47,	/* charset */
				"attributes-charset",
				-1,
				NULL
			},
			{
				0x48,	/* naturalLanguage */
				"attributes-natural-language",
				-1,
				NULL
			},
			{
				0x45,	/* uri */
				"printer-uri",
				-1,
				NULL
			},
			{
				0x41,	/* textWithoutLanguage */
				"status-message",
				-1,
				NULL
			},
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	},
	{
		0x02,	/* job-attributes-tag */
		{
			{
				0x21,	/* integer */
				"job-id",
				-1,
				NULL,
			},
			{
				0x45,	/* uri */
				"job-uri",
				-1,
				NULL,
			},
			{
				0x23,	/* enum */
				"job-state",
				-1,
				NULL,
			},
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	},
	{
		0x04,	/* printer-attributes-tag */
		{
			{
				0x33,	/* range-of-integer */
				"copies-supported",
				-1,
				NULL
			},
			{
				0x47,	/* charset */
				"charset-supported",
				-1,
				NULL
			},
			{
				0x23,	/* enum */
				"printer-state",
				-1,
				NULL
			},
			{
				0x44,	/* keyword */
				"printer-state-reasons",
				-1,
				NULL
			},
			{
				0x22,	/* boolean */
				"printer-is-accepting-jobs",
				-1,
				NULL
			},
			{
				0x49,	/* mimemediaType */
				"document-format-supported",
				-1,
				NULL
			},
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	},
	{
		0x03,	/* end-of-attributes-tag */
		{
			{
				0x03,	/* end-of-attributes */
				NULL,
				-1,
				NULL
			}
		}
	}
};

/*
 *
 */
oppd_signal	sig[] = {
	{SIGCHLD, NULL},
	{-1      , NULL}
};
int		force_abort = 0;

/*
 * signal handler
 */
static
void		sig_handler(
	int		signo
	)
{
	switch (signo) {
		case	SIGCHLD :
			{
				pid_t		pid;
				int		stat;
				while((pid = waitpid(-1, &stat, WNOHANG)) > 0);
			}
			break;
		case	SIGALRM :
			break;
		case	SIGTERM :
			force_abort = 1;
			break;
		default :
			/* ignore */
			break;
	}
	return;
}

/*
 * initialize signals
 */
static
int		sig_init(
	void
	)
{
	int		ecode = 0;
	int		i = 0;
struct	sigaction	act;
struct	sigaction	old;

	while (sig[i].signo != -1) {
		act.sa_handler = sig_handler;
		sigemptyset(&act.sa_mask);
		act.sa_flags = 0;
		if (sigaction(sig[i].signo, &act, &old) < 0) {
			ecode = -1;
			break;
		}
		sig[i].act = malloc(sizeof(struct sigaction));
		if (!sig[i].act) {
			ecode = -1;
			break;
		}
		memcpy(sig[i].act, &old, sizeof(struct sigaction));
		i++;
	}

	return ecode;
}

/*
 * finalize signals
 */
static
void		sig_fin(
	void
	)
{
	int		i = 0;

	while (sig[i].signo != -1) {
		if (!sig[i].act) {
			sigaction(sig[i].signo, sig[i].act, NULL);
			free(sig[i].act);
			sig[i].act = NULL;
		}
		i++;
	}

	return;
}

/*
 * initialize
 */
static
int		init(
	void
	)
{
	int		ecode = 0;

	/* initialize signal */
	if (!ecode) ecode = sig_init();

	/* allocate buffer */
	if (!ecode) {
		buf = malloc(OPPD_BUF_LEN);
		buf_pos = 0;
		if (!buf) ecode = -1;
	}

	memset(printer, 0, sizeof(printer));

	return ecode;
}

/*
 * finalize
 */
static
void		fin(
	void
	)
{
	/* free buffer */
	if (buf) free(buf);

	/* finalize signals */
	sig_fin();

	return;
}

/*
 * clear attributes
 */
static
void		clear_attrib(
	ipp_attrib	att[]
	)
{
	int		i, j;

	for (i = 0; att[i].tag != 0x03; i++) {
		for (j = 0; att[i].attrib[j].tag != 0x03; j++) {
			if (att[i].attrib[j].val) {
				free(att[i].attrib[j].val);
				att[i].attrib[j].len = -1;
				att[i].attrib[j].val = NULL;
			}
		}
	}
}

/*
 * set attribute
 */
static
void		set_attrib(
	ipp_attrib	att[],
	char		*name,
	int		len,
	char		*val
	)
{
	int		i, j;

	for (i = 0; att[i].tag != 0x03; i++) {
		for (j = 0; att[i].attrib[j].tag != 0x03; j++) {
			if (strcasecmp(name, att[i].attrib[j].name)) continue;
			if (len >= 0) {
				if (len) {
					att[i].attrib[j].val
					   = realloc(att[i].attrib[j].val, len);
				} else {
					free(att[i].attrib[j].val);
					att[i].attrib[j].val = NULL;
				}
				if (att[i].attrib[j].val) {
					memcpy(att[i].attrib[j].val, val, len);
				}
				att[i].attrib[j].len = len;
			} else {
				if (att[i].attrib[j].val) {
					free(att[i].attrib[j].val);
					att[i].attrib[j].len = -1;
					att[i].attrib[j].val = NULL;
				}
			}
			break;
		}
	}
}

/*
 * copy attribute
 */
static
void		copy_attrib(
	ipp_attrib	dst[],
	ipp_attrib	src[],
	char		*name
	)
{
	int		i, j;
	char		*val;
	int		len;
	char		a_tag;
	int		found = 0;

	for (i = 0; src[i].tag != 0x03; i++) {
		for (j = 0; src[i].attrib[j].tag != 0x03; j++) {
			if (strcasecmp(name, src[i].attrib[j].name)) continue;
			len = src[i].attrib[j].len;
			val = src[i].attrib[j].val;
			a_tag = src[i].attrib[j].tag;
			found = 1;
			break;
		}
		if (found) break;
	}
	if (!found) return;
	for (i = 0; dst[i].tag != 0x03; i++) {
		for (j = 0; dst[i].attrib[j].tag != 0x03; j++) {
			if (strcasecmp(name, dst[i].attrib[j].name)) continue;
			if (len > 0) {
				dst[i].attrib[j].val
				    = realloc(dst[i].attrib[j].val, len);
			} else {
				free(dst[i].attrib[j].val);
				dst[i].attrib[j].val = NULL;
			}
			if (dst[i].attrib[j].val) {
				memcpy(dst[i].attrib[j].val, val, len);
			}
			dst[i].attrib[j].tag = a_tag;
			dst[i].attrib[j].len = len;
			break;
		}
	}
}

/*
 * get attribute
 */
static
int		get_attrib(
	ipp_attrib	att[],
	char		*name,
	int		*len,
	char		**val
	)
{
	int		ecode = -1;
	int		i, j;

	for (i = 0; att[i].tag != 0x03; i++) {
		for (j = 0; att[i].attrib[j].tag != 0x03; j++) {
			if (strcasecmp(name, att[i].attrib[j].name)) continue;
			if (len) *len = att[i].attrib[j].len;
			if (val) *val = att[i].attrib[j].val;
			ecode = 0;
			break;
		}
	}

	return ecode;
}

/*
 * get attribute string
 */
static
char		*get_attrib_string(
	ipp_attrib	att[],
	char		*name
	)
{
	int		len;
	char		*val;
static	char		str[OPPD_BUF_LEN];

	if (get_attrib(att, name, &len, &val) == 0) {
		if (len < OPPD_BUF_LEN) {
			memcpy(str, val, len);
			str[len] = '\0';
			return str;
		}
	}

	return NULL;
}

/*
 * get request
 */
static
int		get_request(
	void
	)
{
	int		ecode = 0;
	int		start = 0;
	int		ipp = 0;
	int		fl;
	int		len = 0;
static	char		tmp[OPPD_BUF_LEN];
static	char		wrk[OPPD_BUF_LEN];
	int		pos;
	int		result;
	char		*p;

	clear_attrib(req_att);

	fl = fcntl(0, F_GETFL);
	fcntl(0, F_SETFL, fl & ~O_NONBLOCK);

	// while (fgets(tmp, OPPD_BUF_LEN - 1, stdin)) {
	// while ((tmp[0] = fgetc(stdin))) {
	while ((result = read(0, tmp, 1)) == 1) {
		if (tmp[0] == '\r') read(0, tmp, 1);
		for (pos = 0; tmp[pos] != '\n'; read(0, &(tmp[++pos]), 1));
		tmp[++pos] = '\0';
		if (tmp[0] == '\n') break;
		// if (!fgets(&(tmp[1]), OPPD_BUF_LEN - 2, stdin)) break;
		if (start == 0) {
			if (strncmp(tmp, "POST ", 5) == 0) {
				start = 1;
				oppd_strtoken(wrk, &(tmp[5]), " ");
				if ((p = rindex(wrk, '/'))) p++;
				else p = wrk;
				strcpy(printer, p);
			}
		} else {
			if (strncasecmp(tmp,
			                "Content-Type: application/ipp",
			                29) == 0) {
				ipp = 1;
			}
			if (strncasecmp(tmp,
			                "Content-Length: ",
			                16) == 0) {
				len = atoi(&(tmp[16]));
			}
		}
		// if ((strcmp(tmp, "\n") == 0) || (strcmp(tmp, "\r\n") == 0)) {
		if (strcmp(tmp, "\n") == 0) {
			break;
		}
	}

	//fcntl(0, F_SETFL, fl | O_NONBLOCK);

	if (ipp == 0) {
		ecode = -1;
	} else {
		if (read(0, buf, 8) < 8) {
			ecode = -1;
		} else {
			buf_pos = 8;
			ecode = len;
		}
	}

	return ecode;
}

/*
 * send attributes
 */
static
int		make_attrib(
	ipp_attrib	att[],
	char		tag
	)
{
	int		i, j;
	int		ecode = 0;
static	char		tmp[OPPD_BUF_LEN];
	int		pos;
	int		len;

	if (tag == 0x03) {
		memcpy(&(buf[buf_pos]), &tag, 1);
		buf_pos++;
	} else {
		for (i = 0; att[i].tag != 0x03; i++) {
			if (att[i].tag != tag) continue;
			buf[buf_pos++] = tag;
			for (j = 0; att[i].attrib[j].tag != 0x03; j++) {
				if (att[i].attrib[j].val) {
					tmp[0] = att[i].attrib[j].tag;
					len = strlen(att[i].attrib[j].name);
					*(int*)(&(tmp[1])) = htons(len);
					strcpy(&(tmp[3]),
					       att[i].attrib[j].name);
					pos = 3 + len;
					len = att[i].attrib[j].len;
					*(int*)(&(tmp[pos])) = htons(len);
					pos += 2;
					memcpy(&(tmp[pos]),
					       att[i].attrib[j].val, len);
					pos += len;
					if (!ecode) {
						memcpy(&(buf[buf_pos]),
						       tmp, pos);
						buf_pos += pos;
					}
				}
			}
		}
	}

	return ecode;
}

/*
 * reply response
 */
static
int		reply(
	int		req,
	int		stat
	)
{
	int		ecode = 0;
static	char		tmp[OPPD_BUF_LEN];
	char		*msg;
	char		int_buf[8];

	switch (stat) {
		case	0x0000 :
			msg = "successful-ok";
			break;
		case	0x0001 :
			msg = "successful-ok-ignored-or-substituted-attributes";
			break;
		case	0x0002 :
			msg = "successful-ok-conflicting-attributes";
			break;
		case	0x0400 :
			msg = "client-error-bad-request";
			break;
		case	0x0401 :
			msg = "client-error-forbidden";
			break;
		case	0x0404 :
			msg = "client-error-not-possible";
			break;
		case	0x0406 :
			msg = "client-error-not-found";
			break;
		case	0x0500 :
			msg = "server-error-internal-error";
			break;
		case	0x0501 :
			msg = "server-error-operation-not-supported";
			break;
		case	0x0502 :
			msg = "server-error-service-unavailable";
			break;
		case	0x0504 :
			msg = "server-error-device-error";
			break;
		case	0x0505 :
			msg = "server-error-temporary-error";
			break;
		case	0x0506 :
			msg = "server-error-not-accepting-jobs";
			break;
		case	0x0507 :
			msg = "server-error-busy";
			break;
		case	0x0508 :
			msg = "server-error-job-canceled";
			break;
		default :
			stat = 0x0500;
			msg = "server-error-internal-error";
			break;
	}

	*(int*)(&(buf[2])) = htons(stat);

	set_attrib(rep_att, "status-message", strlen(msg), msg);

	msg = "utf-8";
	set_attrib(rep_att, "attributes-charset", strlen(msg), msg);
	set_attrib(rep_att, "charset-supported", strlen(msg), msg);
	msg = "en";
	set_attrib(rep_att, "attributes-natural-language",
	           strlen(msg), msg);
	// copy_attrib(rep_att, req_att, "printer-uri");

	switch (req) {
		case	0x0002 : /* Print-Job */
			if (!ecode) ecode = make_attrib(rep_att, 0x01);
			if (!ecode) ecode = make_attrib(rep_att, 0x03);
			break;
		case	0x0009 : /* Get-Job-Attributes */
			if (!ecode) ecode = make_attrib(rep_att, 0x01);
			if (!ecode) ecode = make_attrib(rep_att, 0x03);
			break;
		case	0x000b : /* Get-Printer-Attributes */
			*(int*)int_buf = htonl(1);
			*(int*)(&(int_buf[4])) = htonl(9999);
			set_attrib(rep_att, "copies-supported", 8, int_buf);
			*(int*)int_buf = htonl(3);
			set_attrib(rep_att, "printer-state", 4, int_buf);
			msg = "none";
			set_attrib(rep_att, "printer-state-reasons",
			           strlen(msg), msg);
			int_buf[0] = 0x01;
			set_attrib(rep_att, "printer-is-accepting-jobs",
			           1, int_buf);
			msg = "appilication/octet-stream";
			set_attrib(rep_att, "document-format-supported",
			           strlen(msg), msg);
			if (!ecode) ecode = make_attrib(rep_att, 0x01);
			if (!ecode) ecode = make_attrib(rep_att, 0x04);
			if (!ecode) ecode = make_attrib(rep_att, 0x03);
			break;
		default :
			if (!ecode) ecode = make_attrib(rep_att, 0x01);
			if (!ecode) ecode = make_attrib(rep_att, 0x03);
			break;
	}

	strcpy(tmp, "HTTP/1.1 200 OK\r\n");
	if (!ecode) if (write(1, tmp, strlen(tmp)) < strlen(tmp)) ecode = -1;
	sprintf(tmp, "Date: %s\r\n", httpGetDateString(time(NULL)));
	if (!ecode) if (write(1, tmp, strlen(tmp)) < strlen(tmp)) ecode = -1;
	strcpy(tmp, "Content-Type: application/ipp;\r\n");
	if (!ecode) if (write(1, tmp, strlen(tmp)) < strlen(tmp)) ecode = -1;
	sprintf(tmp, "Content-Length: %d\r\n\r\n", buf_pos);
	if (!ecode) if (write(1, tmp, strlen(tmp)) < strlen(tmp)) ecode = -1;

	if (!ecode) if (write(1, buf, buf_pos) < buf_pos) ecode = -1;

	return ecode;
}

/*
 * receive attributes
 */
static
int		read_attrib(
	void
	)
{
	int		ecode = -1;
	char		tag = 0;
	char		o_tag;
	char		len_buf[2];
	int		len;
static	char		name[OPPD_BUF_LEN];
static	char		val[OPPD_BUF_LEN];
	int		cnt = 0;

	while (read(0, &tag, 1) == 1) {
		cnt++;
		if (tag == 0x03) {
			ecode = cnt;
			break;
		} else if (tag >= 0x10) {
			/* attribute */
			if (read(0, len_buf, 2) < 2) break;
			cnt += 2;
			len = ntohs(*(int*)len_buf);
			if (read(0, name, len) < len) break;
			cnt += len;
			name[len] = '\0';
			/* value */
			if (read(0, len_buf, 2) < 2) break;
			cnt += 2;
			len = ntohs(*(int*)len_buf);
			if (len > 0) {
				if (read(0, val, len) < len) break;
				cnt += len;
			}
			val[len] = '\0';

			set_attrib(req_att, name, len, val);

		} else {
			o_tag = tag;
		}
	}

	return ecode;
}

/*
 * receive data
 */
static
int		recv_data(
	int		len,
	int 		no_spool
	)
{
	int		ecode = 0;
	int		result;
	int		read_len;
	char		read_buf[OPPD_BUF_LEN];
static	char		spool[OPPD_BUF_LEN];
static	char		title[OPPD_BUF_LEN];
static	char		owner[OPPD_BUF_LEN];
	int		id = 0;
	int		*w;
	int		copies;
	char		*p;
	int		fd = -1;
	oppd_job	job;
	OPPD_ConfEnt   *conf;

	if (!no_spool) {
		if (oppd_getConfEntByName(OPPD_CONF_FILE,printer,&conf) <= 0) {
		    /* printer not found */
		    return -1;
		}
		if (conf->remote) {
		    /* remote printer is not supported */
		    return -1;
		}

		/* id */
		id = oppd_new_job_id(printer);
		/* copies */
		get_attrib(req_att, "copies", NULL, (char **)(&w));
		copies = ntohl(*w);
		if (copies < 1) copies = 1;
		/* title */
		strcpy(title, get_attrib_string(req_att, "job-name"));
		/* owner */
		strcpy(owner,
		       get_attrib_string(req_att, "requesting-user-name"));
		/* spool file */
		fd = oppd_open_spool(printer, &p);
		if (fd < 0) ecode = -1;
		strcpy(spool, p);
	}

	if (!ecode) {
		if (fd >= 0) flock(fd, LOCK_EX);
		while (len) {
			read_len = (OPPD_BUF_LEN < len ? OPPD_BUF_LEN : len);
			result = read(0, read_buf, read_len);
			if (result >= 0) {
				if (!no_spool) {
					if (write(fd,read_buf,result)
					    < result) {
						break;
					}
				}
				len -= result;
			} else {
				break;
			}
		}
		if (len) ecode = -1;
		if (fd >= 0) flock(fd, LOCK_UN);
		if (fd >= 0) close(fd);
		if ((!ecode) && (!no_spool)) {
			job.id     = id;
			job.copies = copies;
			job.title  = title;
			job.owner  = owner;
			job.spool  = spool;
			oppd_add_job(printer, &job);
		}
	}

	return ecode;
}

/*
 * spool
 */
static
int		spool(
	void
	)
{
	int		ecode = 0;
	int		req_id;
	int		fl;
	int		req_len;
	int		result;
	int		read_len;
	int		spool_full;

	while ((req_len = get_request()) >= 0) {
		if (req_len == 0) continue;
		req_len -= 8;
		req_id = ntohs(*(int*)(&(buf[2])));
		if (req_len) {
			result = read_attrib();
			if (result >= 0) {
				req_len -= result;
			} else {
				ecode = result;
				break;
			}
		}
		if (req_id != 0x0002) {
			while (req_len > 0) {
				read_len = (OPPD_BUF_LEN - 8 < req_len ?
				            OPPD_BUF_LEN - 8 : req_len);
				result = read(0, &(buf[8]), read_len);
				if (result > 0) {
					req_len -= result;
				} else if (result < 0) {
					ecode = result;
					break;
				}
			}
		}
		switch (req_id) {
			case	0x0002 : /* Print-Job */
				/* check spool area */
				spool_full = oppd_check_spool_area(printer,
				                                   req_len);
				/* if req_len is 0 then don't spool */
				if (req_len != 0) {
					if (!ecode) ecode = recv_data(req_len,
								      spool_full);
				}
				if (!ecode) {
					if (spool_full) {
						ecode = reply(req_id, 0x0506);
					} else {
						ecode = reply(req_id, 0x0000);
					}
				}
				break;
			case	0x0009 : /* Get-Job-Attributes */
				if (!ecode) ecode = reply(req_id, 0x0000);
				break;
			case	0x000b : /* Get-Printer-Attributes */
				if (!ecode) ecode = reply(req_id, 0x0000);
				break;
			default :
				if (!ecode) ecode = reply(req_id, 0x0501);
				break;
		}
		if (ecode) break;
	}

	if (req_len < 0) ecode = -1;

	return ecode;
}

/*
 * main
 */
int		main(
	int		argc,
	char		**argv
	)
{
	int		ecode = 0;

	/* initialize */
	if (!ecode) ecode = init();

 	/* spool */
	if (!ecode) ecode = spool();

	/* finalize */
	fin();

	return ecode;
}

