/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#include <stdio.h>
#include "oppd_conf.h"
#include <string.h>
#include <ctype.h>

static FILE *confFp = NULL;
static char printerNameBuf[OPPD_MAX_PRINTERNAME+1];
static char deviceNameBuf[OPPD_MAX_DEVICENAME+1];
static char bidiNameBuf[OPPD_MAX_BIDINAME+1];
static char commentBuf[OPPD_MAX_CONFLINE+1];
static OPPD_ConfEnt confEnt = {
    printerNameBuf,0,deviceNameBuf, bidiNameBuf,commentBuf
};

int oppd_openConf(char *fileName)
{
    if (confFp != NULL) fclose(confFp);
    if ((confFp = fopen(fileName,"r")) == NULL) return -1;
    return 0;
}

static char *skipSP(char *p)
{
    int c;

    while ((c = *p) != '\0' && c == ' ' || c == '\t') p++;
    return p;
}

static char *getWord(char *s, char *buf, int buflen)
{
    char *p;
    char *d;
    int c;
    int i;

    p = skipSP(s);
    d = buf;
    for (i = 0;i+2 < buflen && (c = *p) != '\0' && !isspace(c);i++) {
	*d++ = *p++;
    }
    *d = '\0';
    return c == '\n' || c == '\0' ? NULL : p;
}

int oppd_getConfEnt(OPPD_ConfEnt **epp)
{
    char *p;
    char buf[OPPD_MAX_CONFLINE+1];

nextline:
    if (fgets(buf,OPPD_MAX_CONFLINE+1,confFp) == NULL) {
	if (feof(confFp)) return 0; /* EOF */
	return -1; /* read error */
    }
    if (buf[0] == '#') goto nextline; /* skip comment line */

    /* get printer name */
    p = getWord(buf,confEnt.printerName,OPPD_MAX_PRINTERNAME+1);
    if (confEnt.printerName[0] == '\0') {
	return -1; /* no printer name */
    }
    if (p == NULL) return -1; /* no device name */
    p = getWord(p,confEnt.deviceName,OPPD_MAX_DEVICENAME+1);
    if (confEnt.deviceName[0] == '\0') {
	return -1; /* no device name */
    }

    /* check if the printer is remote */
    if (confEnt.deviceName[0] != '/') {
	confEnt.remote = 1;
    } else {
	confEnt.remote = 0;
	if (p == NULL) return -1; /* no bidi name */
	p = getWord(p,confEnt.bidiName,OPPD_MAX_BIDINAME+1);
	if (confEnt.bidiName[0] == '\0') {
	    return -1; /* no bidi name */
	}
    }
    if (p != NULL) {
	int i;
	/* read comment */

	p = skipSP(p);
	for (i = 0;i < OPPD_MAX_CONFLINE && *p != '\0' && *p != '\n';i++) {
	    confEnt.comment[i] = *p++;
	}
	confEnt.comment[i] = '\0';
    } else {
	confEnt.comment[0] = '\0';
    }

    *epp = &confEnt;
    return 1;
}

void oppd_endConfEnt(void)
{
    if (confFp != NULL) {
	fclose(confFp);
	confFp = NULL;
    }
}

int oppd_getConfEntByName(char *fileName, char *printerName, OPPD_ConfEnt **epp)
{
    OPPD_ConfEnt *ep;
    int r;

    if (oppd_openConf(fileName) < 0) return -1;
    while ((r = oppd_getConfEnt(&ep)) > 0) {
	if (strcmp(ep->printerName,printerName) == 0) {
	    *epp = ep;
	    break;
	}
    }
    oppd_endConfEnt();
    return r;
}

int oppd_getConfList(char *fileName, OPPD_ConfEnt **epp)
{
    static OPPD_ConfEnt conf[OPPD_CONF_MAX];
    OPPD_ConfEnt *ep;
    int n = 0;

    memset(conf,0,sizeof(conf));

    if (oppd_openConf(fileName) < 0) return -1;
    while (oppd_getConfEnt(&ep) > 0) {
	if (n >= OPPD_CONF_MAX) return -1; /* too many config entries */
	conf[n].printerName = strdup(ep->printerName);
	conf[n].remote = ep->remote;
	conf[n].deviceName = strdup(ep->deviceName);
	if (!ep->remote) {
	    conf[n].bidiName = strdup(ep->bidiName);
	}
	conf[n].comment = strdup(ep->comment);
	n++;
    }
    oppd_endConfEnt();
    *epp = conf;
    return n;
}
