/*
Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include "opvp_rpc_client.h"
#include "opvp_rpc_core.h"
#include "opvp_rpc_reqno.h"

/* check response packet */
static int checkResponse(void *rpcHandle, int seqNo, int reqNo)
{
    int rreqNo;
    int rseqNo;
    int err = BIDI_OK;

reenter:
    /* get response */
    if ((rseqNo = oprpc_getPktStart(rpcHandle)) < 0) {
	return BIDI_ERROR;
    }
    /* get request number */
    if ((oprpc_getPkt(rpcHandle,(char *)&rreqNo,sizeof(rreqNo))) < 0) {
	return BIDI_ERROR;
    }
    /* check sequence number */
    if (rseqNo != seqNo) {
	err = BIDI_ERROR;
    }
    /* check request number */
    if (rreqNo != reqNo) {
	err = BIDI_ERROR;
    }
    if (err < 0) {
	oprpc_getPktEnd(rpcHandle);
    }
    return err;
}

int CStubBidiStartRead(void *rpcHandle, BidiReadMode idReadMode, char *pLang)
{
    int seqNo;
    int r;

    /* send request */
    if ((seqNo = oprpc_putPktStart(rpcHandle, -1, RPCNO_PDSTARTREAD)) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putPkt(rpcHandle,(char *)&idReadMode,sizeof(idReadMode)) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putStr(rpcHandle,pLang) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }

    /* get response */
    if (checkResponse(rpcHandle,seqNo,RPCNO_PDSTARTREAD) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getInt(rpcHandle,&r) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }
    return r;
}

int CStubBidiRead(void *rpcHandle, void *pBuf, int nBufBytes)
{
    int seqNo;
    int r;
    char *buf;

    /* send request */
    if ((seqNo = oprpc_putPktStart(rpcHandle, -1, RPCNO_PDREAD)) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putInt(rpcHandle,&nBufBytes) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }

    /* get response */
    if (checkResponse(rpcHandle,seqNo,RPCNO_PDREAD) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getInt(rpcHandle,&r) < 0) {
	return BIDI_ERROR;
    }
    if (r > 0) {
	if (oprpc_getPktPointer(rpcHandle,(void **)&buf,r) < 0) {
	    return BIDI_ERROR;
	}
	memcpy(pBuf,buf,r);
    }
    if (oprpc_getPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }
    return r;
}

int CStubBidiEndRead(void *rpcHandle)
{
    int seqNo;
    int r;
    char *buf;

    /* send request */
    if ((seqNo = oprpc_putPktStart(rpcHandle, -1, RPCNO_PDENDREAD)) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }

    /* get response */
    if (checkResponse(rpcHandle,seqNo,RPCNO_PDENDREAD) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getInt(rpcHandle,&r) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }
    return r;
}

int CStubBidiCtrl(void *rpcHandle, int idRequest, void *pdata, int nDataBytes)
{
    int seqNo;
    int r;
    char *buf;

    /* send request */
    if ((seqNo = oprpc_putPktStart(rpcHandle, -1, RPCNO_PDCTRL)) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putInt(rpcHandle,&idRequest) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putInt(rpcHandle,&nDataBytes) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putPktPointer(rpcHandle,pdata,nDataBytes) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_putPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }

    /* get response */
    if (checkResponse(rpcHandle, seqNo,RPCNO_PDCTRL) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getInt(rpcHandle,&r) < 0) {
	return BIDI_ERROR;
    }
    if (oprpc_getPktEnd(rpcHandle) < 0) {
	return BIDI_ERROR;
    }
    return r;
}

