/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include "opvp_rpc_server.h"
#include "opvp_rpc_core.h"
#include "opvp_rpc_reqno.h"

static int SStubBidiStartRead(void *ap, int seqNo)
{
    BidiReadMode idReadMode;
    char *pLang;
    int r;

    /* get parameter */
    if (oprpc_getPkt(ap,(char *)&idReadMode,sizeof(idReadMode)) < 0) {
	return -1;
    }
    if (oprpc_getStr(ap,&pLang) < 0) {
	return -1;
    }
    /* call real proc */
    r = PDBidiStartRead(idReadMode,pLang);

    /* send response */
    if (oprpc_putPktStart(ap,seqNo,RPCNO_PDSTARTREAD) < 0) {
	return -1;
    }
    if (oprpc_putInt(ap,&r) < 0) {
	return -1;
    }
    if (oprpc_putPktEnd(ap) < 0) {
	return -1;
    }
    if (oprpc_flush(ap) < 0) {
	return -1;
    }
    return 0;
}

static int SStubBidiRead(void *ap, int seqNo)
{
    int nBufBytes;
    void *pBuf;
    int r;

    /* get parameter */
    if (oprpc_getInt(ap,&nBufBytes) < 0) {
	return -1;
    }

    if (nBufBytes > 0) {
	if ((pBuf = alloca(nBufBytes)) == NULL) {
	    return -1;
	}
    }
    /* call real proc */
    r = PDBidiRead(pBuf,nBufBytes);

    /* send response */
    if (oprpc_putPktStart(ap,seqNo,RPCNO_PDREAD) < 0) {
	return -1;
    }
    if (oprpc_putInt(ap,&r) < 0) {
	return -1;
    }
    if (r > 0) {
	if (oprpc_putPktPointer(ap,pBuf,r) < 0) {
	    return -1;
	}
    }
    if (oprpc_putPktEnd(ap) < 0) {
	return -1;
    }
    if (oprpc_flush(ap) < 0) {
	return -1;
    }
    return 0;
}

static int SStubBidiEndRead(void *ap, int seqNo)
{
    int r;

    /* no parameter */

    /* call real proc */
    r = PDBidiEndRead();

    /* send response */
    if (oprpc_putPktStart(ap,seqNo,RPCNO_PDENDREAD) < 0) {
	return -1;
    }
    if (oprpc_putInt(ap,&r) < 0) {
	return -1;
    }
    if (oprpc_putPktEnd(ap) < 0) {
	return -1;
    }
    if (oprpc_flush(ap) < 0) {
	return -1;
    }
    return 0;
}

static int SStubBidiCtrl(void *ap, int seqNo)
{
    int nDataBytes;
    void *pdata;
    int idRequest;
    int r;

    /* get parameter */
    if (oprpc_getInt(ap,&idRequest) < 0) {
	return -1;
    }
    if (oprpc_getInt(ap,&nDataBytes) < 0) {
	return -1;
    }
    if (oprpc_getPktPointer(ap,&pdata,nDataBytes) < 0) {
	return -1;
    }

    /* call real proc */
    r = PDBidiCtrl(idRequest,pdata,nDataBytes);

    /* send response */
    if (oprpc_putPktStart(ap,seqNo,RPCNO_PDCTRL) < 0) {
	return -1;
    }
    if (oprpc_putInt(ap,&r) < 0) {
	return -1;
    }
    if (oprpc_putPktEnd(ap) < 0) {
	return -1;
    }
    if (oprpc_flush(ap) < 0) {
	return -1;
    }
    return 0;
}

int oprpc_doLoop(void *ap)
{
    int reqNo;
    int reqSeqNo;

    while ((reqSeqNo = oprpc_getPktStartNonBlock(ap)) > 0) {
	if (oprpc_getPkt(ap,(char *)&reqNo,sizeof(reqNo)) < 0) {
	    return -1;
	}
	switch (reqNo) {
	case RPCNO_PDSTARTREAD:
	    if (SStubBidiStartRead(ap,reqSeqNo) < 0) {
		return -1;
	    }
	    break;
	case RPCNO_PDREAD:
	    if (SStubBidiRead(ap,reqSeqNo) < 0) {
		return -1;
	    }
	    break;
	case RPCNO_PDENDREAD:
	    if (SStubBidiEndRead(ap,reqSeqNo) < 0) {
		return -1;
	    }
	    break;
	case RPCNO_PDCTRL:
	    if (SStubBidiCtrl(ap,reqSeqNo) < 0) {
		return -1;
	    }
	    break;
	case RPCNO_ECHO:
	case RPCNO_READY:
	default:
	    return -1;
	}
	if (oprpc_getPktEnd(ap) < 0) {
	    return -1;
	}
    }
    return reqSeqNo;
}
